/* t_quant.c -- an algorithm for quantization of discrete probability distributions */
/* 09/04/2011, written by Yuriy A. Reznik <yreznik@ieee.org> */

/* This software and documentation is provided "as is", and the author makes 
 * no representations or warranties, expressed or implied, including, but not
 * limited to, warranties of merchantability or fitness for any particular
 * purpose, or that the use of this software or documentation will not infringe 
 * any third parties patents, copyrights, trademarks or other rights. The 
 * author will not be liable for any direct, indirect, special or consequential 
 * damages arising out of any use of the software or documentation. */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "t_quant.h"

/* limits: */
#define M_MAX   20    /* max number of dimensions */
#define N_MAX   33    /* max value of common denominator */

/* Pascal's triangle: */
static int *binomial[N_MAX+1], b_data[(N_MAX+1) * (N_MAX+2) / 2];  

/*
 * Initialize quantization module.
 */
void t_init ()
{
  int n, k, *b = b_data;
  for (n=0; n<=N_MAX; n++) {
    binomial[n] = b; b += n + 1;            /* row pointer */
    binomial[n][0] = binomial[n][n] = 1;    /* 1st & last coeffs */
    for (k=1; k<n; k++)                     /* compute coeffs in the middle */
      binomial[n][k] = binomial[n-1][k-1] + binomial[n-1][k];
  }
}

/*
 * Compute number of reconstruction points. 
 * Input:
 *   m - number of dimensions
 *   n - precision parameter
 * Returns: 
 *   number of types with m dimensions and denominator n
 */
int t_points (int m, int n)
{
  assert(m < M_MAX);  
  assert(n < N_MAX);  
  assert(n+m-1 < N_MAX);  
  assert(binomial[0][0] == 1); 
  /* return number of types: */
  return binomial[n+m-1][m-1];
}

/* callback for qsort(): */
typedef struct {float delta; int i;} D;
int D_cmp (const D *a, const D *b) 
{
  return (a->delta > (b->delta)? 1: ((a->delta < b->delta)? -1: 0));
}

/*
 * Quantize probability distribution.
 * Input:
 *   p[0..m-1] - distribution to quantize
 *   n -  precision parameter
 * Returns:
 *   index of nearest lattice point
 */
int t_quant (int m, int n, float *p)
{
  int i, j, idx, n1, Delta, s; 
  int k[M_MAX]; D d[M_MAX]; 

  /* quantize distribution p[] to type k[]: */
  for (n1=0,i=0; i<m; i++) n1 += (k[i] = (int) floor(p[i] * n + 0.5));     // unconstrained quantization
  Delta = n1 - n;
  if (Delta != 0) {
    for (i=0; i<m; i++) {d[i].delta = (float)k[i] - p[i] * n; d[i].i = i;} // compute & sort errors
    qsort(d, m, sizeof(D), D_cmp);      
    if (Delta > 0) {for (j=m-Delta; j<m; j++) k[d[j].i] --;}               // ensure that sum_i k[i] = n
    else           {for (j=0; j<abs(Delta); j++) k[d[j].i] ++;}
  }

  /* compute type index: */
  for (i=0, idx=0; i<m-2; i++) {
    s = 0;
    for (j=0; j<k[i]; j++)
      s += binomial[n-j+m-i-2][m-i-2];
    idx += s;
    n -= k[i];
  }
  idx += k[m-2];

  return idx;
}

/*
 * Reconstruct quantized probability distribution. 
 * Input:
 *   m - number of dimensions, n - precision parameter
 *   idx - index of a reconstruction point
 * Output:
 *   p[0..m-1] - reconstructed probability distribution
 * Returns:
 *   1 - success; 0 - invalid index
 */
int t_reconst (int m, int n, int idx, float *p)
{
  int i, j, k[M_MAX], s, x; float n_inv = 1.f/(float)n; 

  /* check index: */
  if (idx < 0 || idx >= t_points(m,n))
    return 0;

  /* decode type: */
  for (i=0; i<m-2; i++) {
    s = 0;
    for (j=0; j<n; j++) {
      x = binomial[n-j+m-i-2][m-i-2];
      if (idx - s < x) break;
      s += x;               
    }
    k[i] = j;
    idx -= s;
    n -= j;
  }
  k[m-2] = idx;
  k[m-1] = n - idx;

  /* convert it to distribution: */
  for (j=0; j<m; j++) 
    p[j] = (float)k[j] * n_inv;

  return 1;
}


/********************** TEST PROGRAMS: *************************/
#if TEST_ACCURACY 

/* L1 distance: */
double dist (int m, float *p1, float *p2)
{ 
  double s = 0.; int i;
  for (i=0; i<m; i++) s += fabs(p1[i] - p2[i]);
  return s;
}

/* test program: */
int main ()
{
  float p[M_MAX], q[M_MAX]; // original & quantized distributions 
  int a, i, j, m=3, n, n0 = 27, t, t0;
  double e, e1; 

  t_init ();
  t0 = t_points(m,n0);
  a = m/2;

  /* try different lattice parameters: */
  for (n=1; n<20; n++) {

    /* report # of points & rate: */
    t = t_points(m,n);
    printf("n=%d, t=%d, rate=%g, ", n, t, log((double)t) / log(2.));

    /* scan finer lattice (n0): */
    e = 0.;
    for (i=0; i<t0; i++) {
      /* pick distribution: */
      t_reconst (m, n0, i, p);
      /* quantize it using given lattice parameter (n): */
      j = t_quant (m, n, p);
      t_reconst (m, n, j, q);
      /* compute worst case quantization error: */
      e1 = dist (m, p, q);
      if (e1 > e) e = e1;
    }

    /* report measured quant. error and compare it with known value for L1 radius: */
    printf("L1 error = %g vs. %g\n", e, (double)2*a*(m-a) / (double)(m*n) );
  }

  return 0;
}
#endif

#ifdef TEST_SPEED
#if defined(_WIN32) || defined(_WIN64)
#include <windows.h> 

/* test program: */
int main ()
{
  float p[M_MAX], q[M_MAX];  // original & quantized distributions 
  static struct {int m, n, n0;} modes[] = {{8, 6, 24}, {16, 8, 10}}; // modes to test
  int nmodes = sizeof (modes) / sizeof(modes[0]);
  long long freq, start_time, end_time, quant_time, reconst_time;
  int i, j, k, m, n, n0, t0;

  /* init library: */
  t_init ();

  /* limit execution to 1 thread, get CPU frequency: */
  SetThreadAffinityMask (GetCurrentThread(), 1);
  QueryPerformanceFrequency ((LARGE_INTEGER*)&freq);
  printf("CPU frequency: %g [GHz]\n", (double)freq / 1.0e9);

  printf("\nTesting speed of type quantization:\n");

  /* try different modes: */
  for (i=0; i<nmodes; i++) 
  {
    m = modes[i].m; n = modes[i].n; n0 = modes[i].n0;
    t0 = t_points(m,n0);    

    /* say what we are testing: */ 
    printf("\nm=%d, n=%d, test points=%d:\n", m, n, t0);

    /* scan points: */
    quant_time = reconst_time = 0;
    for (j=0; j<t0; j++) {

      /* pick a distribution: */
      t_reconst (m, n0, j, p);

      /* quantize it: */
      QueryPerformanceCounter ((LARGE_INTEGER*)&start_time);
      k = t_quant (m, n, p);
      QueryPerformanceCounter ((LARGE_INTEGER*)&end_time);
      quant_time += end_time - start_time;

      /* compute a reconstruction point: */
      QueryPerformanceCounter ((LARGE_INTEGER*)&start_time);
      t_reconst (m, n, k, q);
      QueryPerformanceCounter ((LARGE_INTEGER*)&end_time);
      reconst_time += end_time - start_time;
    }

    /* report results: */
    printf("Average quantization time: %g [cycles], %g [secs]\n", (double)quant_time /(double)t0, (double)quant_time / ((double)t0 * freq));
    printf("Average reconstruction time: %g [cycles], %g [secs]\n", (double)reconst_time /(double)t0, (double)reconst_time / ((double)t0 * freq));
  }

  return 0;
}

#endif
#endif